from typing import Optional

from .resolver import Resolver
from .scope_manager import scope_manager
from .scopetype import parse_scope_type


def ensure_scope(scope_id: str = 'root', scope_type: str = 'root', parent_id: Optional[str] = None) -> None:
    if parent_id is None and scope_id != 'root':
        parent_id = 'root'
    return scope_manager.ensure_scope(scope_id, parse_scope_type(scope_type), parent_id)


def add_context(scope_id: str = 'root', **kwargs) -> None:
    scope_manager.add_context(scope_id, kwargs)


def get_resolver(scope_id: str = 'root') -> Resolver:
    return scope_manager.get_resolver(scope_id)


def clear_scope(scope_id: str) -> None:
    scope_manager.clear_scope(scope_id=scope_id)


def reset() -> None:
    scope_manager.reset()



from __future__ import annotations

from typing import TypeVar, Type, Callable, Any, Optional

T = TypeVar('T')


class Resolver:
    def __init__(self, method: Callable[[Type, Optional[str], dict[str, str]], Any]):
        self._method = method

    def __call__(self, dependency_type: Type[T], id=None) -> T:
        return self._method(dependency_type, id)




from abc import ABC, abstractmethod
from dataclasses import dataclass


class ScopeType(ABC):
    @abstractmethod
    def supports_storing(self) -> bool:
        pass


@dataclass(frozen=True)
class RootType(ScopeType):
    def supports_storing(self) -> bool:
        return True


@dataclass(frozen=True)
class Transient(ScopeType):
    def supports_storing(self) -> bool:
        return False


@dataclass(frozen=True)
class CustomScope(ScopeType):
    scope_type: str

    def supports_storing(self) -> bool:
        return True


def parse_scope_type(value: str) -> ScopeType:
    if value == 'root':
        return RootType()
    if value == 'transient':
        return Transient()

    return CustomScope(value)



from __future__ import annotations

from typing import Type, TypeVar, Optional

from .dependency_bag import DependencyBag
from .dependency import ConcreteDependency, Dependency
from .composer import Composer
from .resolver import Resolver
from .scope_context import ScopeContext
from .scopetype import ScopeType, RootType, Transient

T = TypeVar('T')


class Scope:
    def __init__(
        self,
        scope_id: str,
        scope_type: ScopeType,
        parent: Optional[Scope] = None
    ):
        self._scope_id = scope_id
        self._parent = parent
        self._scope_type = scope_type
        self._dependency_bag: DependencyBag = DependencyBag.empty()
        self._context: ScopeContext = ScopeContext.empty()
        self._validate()

    def _validate(self):
        if self._scope_id == 'root' and self._scope_type != RootType():
            raise Exception('Only root scope can have root type')
        if self._scope_id != 'root' and self._scope_type == RootType():
            raise Exception('Only root scope can have root type')
        if self._scope_id == 'root' and self._parent is not None:
            raise Exception('Root scope can not have parent')
        if self._scope_type in self.get_ancestor_types():
            raise Exception('Ancestor already has a scope of this type')
        if self._scope_type == Transient():
            raise Exception('Scopes can not have type transient')

    def get_id(self) -> str:
        return self._scope_id

    def get_parent(self) -> Scope:
        return self._parent

    def get_type(self) -> ScopeType:
        return self._scope_type

    def __hash__(self) -> int:
        return hash(self._scope_id)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Scope):
            return False

        return self._scope_id == other.get_id()

    def get_ancestors(self) -> set[Scope]:
        if self._parent is None:
            return set()

        return {self._parent} | self._parent.get_ancestors()

    def get_ancestor_types(self) -> set[ScopeType]:
        return {scope.get_type() for scope in self.get_ancestors()}

    def add_composers(self, composers: list[Composer]) -> None:
        for composer in composers:
            if composer.scope_type.supports_storing() and composer.scope_type != self._scope_type:
                continue

            self._dependency_bag = self._dependency_bag.add(composer)

    def add_context(self, values: dict[str, str]) -> None:
        self._context = self._context.add(values)

    def get_context(self) -> ScopeContext:
        return self._context

    def resolver(self) -> Resolver:
        return Resolver(self.get)

    def _get_from_dependency(self, dependency: Dependency, resolver: Resolver):
        if isinstance(dependency, ConcreteDependency):
            return dependency.value
        if isinstance(dependency, Composer):
            concrete = dependency.create(resolver)
            if dependency.scope_type.supports_storing():
                self._dependency_bag = self._dependency_bag.add(concrete)
            return concrete.value

        raise Exception

    def get(self, dependency_type: Type[T], dependency_id: Optional[str]) -> T:
        result = self._dependency_bag.find(dependency_type, dependency_id, self._context.values())
        if result is not None:
            return self._get_from_dependency(result, self.resolver())

        if self._parent is None:
            raise Exception(f'No candidate for type {dependency_type}')

        return self._parent.get(dependency_type, dependency_id)



import builtins
import inspect
from typing import Type, Optional

from tidipy import Resolver
from tidipy.composer import Composer
from tidipy.conditions import Condition, Conditions
from tidipy.scope_manager import scope_manager
from tidipy.scopetype import parse_scope_type


class AutoFactory:
    def __init__(self, dependency_type: Type):
        self._dependency_type = dependency_type

    def __call__(self, resolver: Resolver):
        constructor = self._dependency_type.__init__
        sig = inspect.signature(constructor)

        params = list(sig.parameters.values())[1:]

        args = []
        for param in params:
            if param.annotation is inspect.Parameter.empty:
                raise TypeError(f"Cannot resolve untyped parameter: {param.name}")
            dependency = resolver(param.annotation)
            args.append(dependency)

        return self._dependency_type(*args)


def auto_compose(
    dependency_type: Type,
    *,
    id: Optional[str] = None,
    scope_type: str = 'root',
    **kwargs
):
    if not inspect.isclass(dependency_type):
        raise ValueError
    factory = AutoFactory(dependency_type)
    composer = Composer(
            id=str(builtins.id(dependency_type)) if id is None else id,
            scope_type=parse_scope_type(scope_type),
            conditions=Conditions(
                conditions={
                    Condition(
                        key=key,
                        one_of_values=frozenset(value) if isinstance(value, set) else frozenset({value})
                    )
                    for key, value in kwargs.items()
                }
            ),
            dependency_type=dependency_type,
            factory=factory
        )
    scope_manager.add_composers([composer])



from __future__ import annotations

import builtins
import inspect
from dataclasses import dataclass
from typing import TypeVar, Callable, Type, get_type_hints, Optional

from .conditions import Conditions, Condition
from .dependency import Dependency, ConcreteDependency
from .resolver import Resolver
from .scopetype import ScopeType, parse_scope_type

T = TypeVar('T')

FactoryMethod = Callable[[Resolver], T]


@dataclass(frozen=True)
class Composer(Dependency):
    id: str
    scope_type: ScopeType
    conditions: Conditions
    factory: Callable[[Resolver], T]
    dependency_type: Type

    def get_id(self) -> str:
        return self.id

    def get_dependency_type(self) -> Type:
        return self.dependency_type

    def get_conditions(self) -> Conditions:
        return self.conditions

    def create(self, resolver: Resolver) -> ConcreteDependency:
        return ConcreteDependency(
            id=self.id,
            conditions=self.conditions,
            value=self.factory(resolver)
        )


def composer(
    factory: Optional[Callable] = None,
    *,
    id: Optional[str] = None,
    scope_type: str = 'root',
    **kwargs: str | set[str]
):
    parsed_scope_type = parse_scope_type(scope_type)

    def inner(func: Callable):
        has_parameter = len(inspect.signature(func).parameters) > 0
        return Composer(
            id=str(builtins.id(func)) if id is None else id,
            scope_type=parsed_scope_type,
            conditions=Conditions(
                conditions={
                    Condition(
                        key=key,
                        one_of_values=frozenset(value) if isinstance(value, set) else frozenset({value})
                    )
                    for key, value in kwargs.items()
                }
            ),
            factory=func if has_parameter else lambda resolve: func(),
            dependency_type=get_type_hints(func).get('return', object)
        )

    if factory is not None:
        return inner(factory)

    return inner



from dataclasses import dataclass


@dataclass(frozen=True)
class Condition:
    key: str
    one_of_values: set[str]

    def is_fulfilled_by(self, value_map: dict[str, str]) -> bool:
        if self.key not in value_map:
            return True

        return value_map[self.key] in self.one_of_values


@dataclass(frozen=True)
class Conditions:
    conditions: set[Condition]

    def is_fulfilled_by(self, value_map: dict[str, str]) -> bool:
        return all(condition.is_fulfilled_by(value_map) for condition in self.conditions)



import importlib
import inspect
from pathlib import Path

from .composer import Composer
from .scope_manager import scope_manager


def scan(package):
    composers = []
    package_path = Path(package.__file__).parent
    package_name = package.__name__

    def walk_modules(path, prefix):
        for item in path.iterdir():
            if item.is_dir():
                yield from walk_modules(item, f"{prefix}.{item.name}")
            elif item.suffix == '.py' and item.name != '__init__.py':
                modname = f"{prefix}.{item.stem}"
                try:
                    mod = importlib.import_module(modname)
                    yield mod
                except Exception:
                    pass

    # Add the package module itself
    try:
        mod = importlib.import_module(package_name)
        yield_mods = [mod]
    except Exception as e:
        print(f"Error importing {package_name}: {e}")
        yield_mods = []

    # Add all submodules
    yield_mods += list(walk_modules(package_path, package_name))

    for mod in yield_mods:
        for name, obj in inspect.getmembers(mod):
            if isinstance(obj, Composer):
                composers.append(obj)

    scope_manager.add_composers(composers)



from __future__ import annotations

from typing import Any, Optional, TypeVar, Generic, Type

from .dependency import Dependency

T = TypeVar('T', bound=Dependency)


class DependencyBag:
    def __init__(self, dependencies: dict[str, Dependency]):
        self._dependencies = dependencies

    @staticmethod
    def empty() -> DependencyBag:
        return DependencyBag({})

    def add(self, dependency: Dependency) -> DependencyBag:
        return DependencyBag(
            {
                **self._dependencies,
                dependency.get_id(): dependency
            }
        )

    def find(
        self,
        dependency_type: Type,
        dependency_id: Optional[str],
        filter_values: dict[str, str]
    ) -> Optional[Dependency]:
        candidates = [
            dependency
            for dependency in self._dependencies.values()
            if issubclass(dependency.get_dependency_type(), dependency_type)
            and dependency.get_conditions().is_fulfilled_by(filter_values)
        ]
        if dependency_id is not None:
            candidates = [dependency for dependency in candidates if dependency.get_id() == dependency_id]
        if len(candidates) == 1:
            return next(iter(candidates))
        if len(candidates) > 1:
            raise Exception(f'More than 1 candidate for type {dependency_type}')

        return None



from typing import Optional

from .composer import Composer
from .scope import Scope
from .scopetype import ScopeType, RootType
from .resolver import Resolver


class ScopeManager:
    def __init__(self):
        self._composers = []
        self._scopes: dict[str, Scope] = {}
        self.ensure_scope(scope_id='root', scope_type=RootType())

    def add_composers(self, composers: list[Composer]) -> None:
        self._composers = self._composers + composers
        for scope in self._scopes.values():
            scope.add_composers(composers)

    def get_resolver(self, scope_id: str) -> Resolver:
        return self._scopes[scope_id].resolver()

    def ensure_scope(self, scope_id: str, scope_type: ScopeType, parent_id: Optional[str] = None) -> None:
        if scope_id not in self._scopes:
            self._create_scope(scope_id, scope_type, parent_id)
            return

        existing_scope = self._scopes[scope_id]
        if existing_scope.get_type() != scope_type:
            raise Exception
        if existing_scope.get_parent() is None and parent_id is not None:
            raise Exception
        if existing_scope.get_parent() is not None and existing_scope.get_parent().get_id() != parent_id:
            raise Exception

    def _create_scope(self, scope_id: str, scope_type: ScopeType, parent_id: Optional[str]) -> None:
        parent = self._scopes[parent_id] if parent_id is not None else None
        scope = Scope(
            scope_id=scope_id,
            parent=parent,
            scope_type=scope_type
        )
        scope.add_composers(self._composers)
        if parent is not None:
            scope.add_context(parent.get_context().values())

        if scope_id in self._scopes:
            raise Exception('Scope already exists')

        self._scopes[scope_id] = scope

    def add_context(self, scope_id: str, values: dict[str, str]):
        self._scopes[scope_id].add_context(values)

    def clear_scope(self, scope_id: str):
        deleted_scope = self._scopes.pop(scope_id)
        child_ids = {scope.get_id() for scope in self._scopes.values() if scope.get_parent() == deleted_scope}
        for child_id in child_ids:
            self.clear_scope(child_id)

        if scope_id == 'root':
            self.ensure_scope(scope_id='root', scope_type=RootType())

    def reset(self) -> None:
        self.__init__()


scope_manager = ScopeManager()



from .composer import composer
from .resolver import Resolver
from .scan import scan
from .scope_manager_api import ensure_scope, clear_scope, reset, get_resolver, add_context
from .autocompose import auto_compose





from __future__ import annotations


class ScopeContext:
    def __init__(self, values: dict[str, str]):
        self._values = values

    @staticmethod
    def empty() -> ScopeContext:
        return ScopeContext({})

    def add(self, values: dict[str, str]) -> ScopeContext:
        if not set(values.keys()).isdisjoint(set(self._values.keys())):
            raise Exception('Can only add values to scope context, not override values')

        return ScopeContext({**self._values, **values})

    def values(self) -> dict[str, str]:
        return self._values



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Type, Any

from .conditions import Conditions


class Dependency(ABC):
    @abstractmethod
    def get_id(self) -> str:
        pass

    @abstractmethod
    def get_dependency_type(self) -> Type:
        pass

    @abstractmethod
    def get_conditions(self) -> Conditions:
        pass


@dataclass(frozen=True)
class ConcreteDependency(Dependency):
    id: str
    value: Any
    conditions: Conditions

    def get_id(self) -> str:
        return self.id

    def get_dependency_type(self) -> Type:
        return type(self.value)

    def get_conditions(self) -> Conditions:
        return self.conditions
